#############################################################################
#
#  Copyright (c) 2022-2023 Paulo Moura  
#  Copyright (c) 2022 Anne Brecklinghaus, Michael Leuschel, dgelessus
#  SPDX-License-Identifier: MIT
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################


"""
Default SWI-Prolog kernel implementation.

Defines the inspection for other predicates than the ones defined in the module jupyter.
"""


from logtalk_kernel.logtalk_kernel_base_implementation import LogtalkKernelBaseImplementation

# When overriding this at a different location, add the location of the logtalk_kernel_base_implementation.py file defining the LogtalkKernelBaseImplementation class to the search path for modules
#sys.path.append('/path/to/kernel/logtalk_kernel/logtalk_kernel')
#from logtalk_kernel_base_implementation import LogtalkKernelBaseImplementation


class LogtalkKernelImplementation(LogtalkKernelBaseImplementation):

    def do_inspect(self, code, cursor_pos, detail_level=0, omit_sections=()):
        """
        For SWI-Prolog, help for a predicate can be accessed with help/1.
        When inspecting a token, the output of this predicate precedes the docs for predicates from module jupyter.
        """
        # Get the matching predicates from module jupyter
        token, jupyter_data = self.get_token_and_jupyter_predicate_inspection_data(code, cursor_pos)

        if not token:
            # There is no token which can be inspected
            return {'status': 'ok', 'data': {}, 'metadata': {}, 'found': False}

        try:
            # Request predicate help with help/1
            response_dict = self.server_request(0, 'call', {'code':'help(' + token + ')'})
            help_output = response_dict["result"]["1"]["output"]

        except Exception as exception:
            self.logger.error(exception, exc_info=True)
            help_output = ''

        found = True

        if help_output == '':
            # There is no help/1 ouput
            if jupyter_data == {}:
                data = {}
                found = False
            else:
                data = jupyter_data
        else:
            # There is help/1 ouput
            jupyter_docs_plain = help_output
            jupyter_docs_md = '<pre>' + help_output.replace('\n', '<br>').replace('$', '&#36;') + '</pre>'

            if jupyter_data != {}:
                # Append the jupyter docs
                jupyter_docs_plain += '\n\n' + '_'*80 + '\n\n' + jupyter_data['text/plain']
                jupyter_docs_md += '<br>' + '_'*80 + '<br><br>' + jupyter_data['text/markdown']

            data = {'text/plain': jupyter_docs_plain, 'text/markdown': jupyter_docs_md}

        return {'status': 'ok', 'data': data, 'metadata': {}, 'found': found}



#############################################################################
#
#  Copyright (c) 2022-2023 Paulo Moura  
#  Copyright (c) 2022 Anne Brecklinghaus, Michael Leuschel, dgelessus
#  SPDX-License-Identifier: MIT
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################


"""
Base class for the actual execution of requests received by the Logtalk Jupyter kernel.
It provides code for starting a Logtalk server and communicating with it.
Additionally, code completion and inspection are implemented.

By subclassing this, basically all functionality of the Logtalk kernel can be overriden.
For further information, see 'kernel.py'.
"""


import json
import logging
import os
import platform
import subprocess

from graphviz import render
from IPython.core.completer import CompletionSplitter
from IPython.utils.tokenutil import line_at_cursor
from os import remove
from signal import signal, SIGINT


path = os.path.dirname(__file__)


class LogtalkKernelBaseImplementation:

    error_ansi_escape_codes =  "\x1b[1;31m" # red and bold

    def __init__(self, kernel):
        self.kernel = kernel

        self.logger = kernel.logger
        self.logger.setLevel(logging.DEBUG)

        self.backend = kernel.backend
        self.backend_data = kernel.active_backend_data

        self.logtalk_proc = None
        self.is_server_restart_required = False

        # Run handle_signal_interrupt when the kernel is interrupted
        signal(SIGINT, self.handle_signal_interrupt)

        # Start the Logtalk server
        self.start_logtalk_server()

        self.configure_token_splitters()
        #self.retrieve_predicate_information()


    def start_logtalk_server(self):
        """Tries to (re)start the Logtalk server process with the configured arguments."""
        # Check if the Logtalk server is to be started with the default program arguments
        # Otherwise, the provided path needs to be absolute or relative to the current working directory
        program_arguments = self.backend_data["program_arguments"]
        if program_arguments == "default":
            # Use the default
            program_arguments = self.kernel.default_program_arguments[self.backend]
            # The third element of the list is the path to the Logtalk source code relative to the directory this file is located in
            # In order for it to be found, the path needs to be extended by the location of this file
            program_arguments[3] = program_arguments[3].replace("logtalk_server/loader.lgt", os.path.join(path, os.path.join("logtalk_server", "loader.lgt")))
            program_arguments[3] = program_arguments[3].replace("\\", "\\\\")

        # Log the program arguments and the directory from which the program is tried to be started
        self.logger.debug('Trying to start the Logtalk server from ' + str(os.getcwd()) + ' with arguments: ' + str(program_arguments))

        # Kill the running Logtalk server
        self.kill_logtalk_server()
        self.logtalk_proc = None

        # Start the Logtalk server
        if platform.system() == 'Windows':
            extended_program_arguments = ["pwsh.exe", "-ExecutionPolicy", "Unrestricted", "-Command"] + program_arguments
        else:
            extended_program_arguments = program_arguments
        self.logtalk_proc = subprocess.Popen(
            extended_program_arguments,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            encoding='UTF-8'
        )
        self.logger.debug(str(self.logtalk_proc))

        # Test if the server was started correctly by requesting the Prolog backend identifier
        try:
            # In case of SICStus Prolog, if the implementation is started with a file which does not exist, no response can be read
            # The kernel cannot stop from trying to read a response and therefore cannot output an error message
            backend_response_dict = self.server_request(0, 'backend', log_response=False)
            self.logger.debug("Started the Logtalk server with the '" + backend_response_dict["result"] + "' backend")
            self.is_server_restart_required = False
            # If logging is configured for the server, send a request to create a log file and thereby enable logging
            if self.kernel.server_logging == True:
                logging_response = self.server_request(0, 'enable_logging', log_response=False)
                if logging_response == 'false':
                    self.logger.debug('No log file could be created by the Logtalk server')
        except Exception as exception:
            raise Exception("The Logtalk server could not be started with the arguments " + str(program_arguments))


    def handle_signal_interrupt(self, signal_received, frame):
        self.handle_interrupt()


    def handle_interrupt(self):
        # Interrupting the kernel interrupts the running Prolog processes, so all of them need to be restarted
        self.kernel.interrupt_all()


    def kill_logtalk_server(self):
        """Kills the Logtalk server process if it is still running."""
        if self.logtalk_proc is not None:
            self.logger.debug(self.backend + ': Kill Logtalk server')
            self.logtalk_proc.kill()
            self.is_server_restart_required = True


    def configure_token_splitters(self):
        """Configures splitters which are used to determine the token which is to be completed/inspected."""
        splitter_delims = ' ,\t\n()[{]}|;\'"'
        self.completion_splitter = CompletionSplitter()
        self.completion_splitter.delims = splitter_delims
        # For the inspection additionally use ':' as a delimiter for splitting as most of the predicate names the tokens are compared to are not module name expanded
        self.inspection_splitter = CompletionSplitter()
        self.inspection_splitter.delims = splitter_delims + ':'


    def retrieve_predicate_information(self):
        """Requests information from the Logtalk server which is needed for code completion and inspection."""
        try:
            # The currently defined predicates are used for code completion
            response_dict = self.server_request(0, 'call', {'code':'jupyter::update_completion_data.'}, log_response=False)
            self.current_predicates = response_dict['result']['1']['predicate_atoms']

            # Retrieve the documentation texts which are shown when a predicate provided by the Logtalk server in the module 'jupyter' is inspected
            jupyter_predicate_docs_dict = self.server_request(0, 'jupyter_predicate_docs', log_response=False)
            self.jupyter_predicate_docs = jupyter_predicate_docs_dict["result"]

        except Exception as exception:
            self.logger.error(exception, exc_info=True)


    ############################################################################
    # Overriden kernel methods
    ############################################################################


    def do_shutdown(self, restart):
        self.kill_logtalk_server()
        return {'status': 'ok', 'restart': restart}


    def do_execute(self, code, silent, store_history=True, user_expressions=None, allow_stdin=False):
        """
        A request to execute code was received.
        The code is tried to be executed by sending it to the Logtalk server.

        If the execution is interrupted or an exception occurs, an error response is sent to the frontend.
        """
        if not silent:
            error_prefix = self.backend_data["error_prefix"]
            try:
                # Check if the server had been shutdown (because of 'halt', an interrupt, or an exception) and a server restart is necessary
                if self.is_server_restart_required:
                    self.logger.debug(self.backend + ': Restart Logtalk server')
                    self.start_logtalk_server()
                    self.send_response_display_data(self.backend_data["informational_prefix"] + 'The Logtalk server was restarted', self.error_ansi_escape_codes)

                # Send an execution request and handle the response
                response_dict = self.server_request(self.kernel.execution_count, 'call', {'code':code})
                if 'result' in response_dict:
                    reply_object = self.handle_success_response(response_dict)
                else:
                    # 'error' in response_dict:
                    reply_object = self.handle_error_response(response_dict)

            except KeyboardInterrupt:
                self.handle_interrupt()
                return {'status': 'error', 'ename' : 'interrupt', 'evalue' : '', 'traceback' : ''}
            except BrokenPipeError:
                self.logger.error(error_prefix + 'Broken pipe\n' + error_prefix + 'The Logtalk server needs to be restarted', self.error_ansi_escape_codes)
                self.is_server_restart_required = True
                self.send_response_display_data(error_prefix + 'Something went wrong\n' + error_prefix + 'The Logtalk server needs to be restarted\n', self.error_ansi_escape_codes)
                return {'status': 'error', 'ename' : 'broken pipe', 'evalue' : '', 'traceback' : ''}
            except Exception as exception:
                self.logger.error(exception, exc_info=True)
                self.is_server_restart_required = True
                self.send_response_display_data(error_prefix + 'Something went wrong\n' + error_prefix + 'The Logtalk server needs to be restarted\n', self.error_ansi_escape_codes)
                return {'status': 'error', 'ename' : 'exception', 'evalue' : '', 'traceback' : ''}
        else:
            reply_object = {
                'status': 'ok',
                'execution_count': self.kernel.execution_count,
                'payload': [],
                'user_expressions': {},
            }

        return reply_object


    def do_complete(self, code, cursor_pos):
        if self.current_predicates is None:
            return {'matches' : [],
                    'cursor_end' : cursor_pos,
                    'cursor_start' : cursor_pos,
                    'metadata' : {},
                    'status' : 'ok'}

        token = self.get_current_token(self.completion_splitter, code, cursor_pos)

        # Find the matching predicates
        # If a key of the dictionary contains the current token, the element is assumed to match
        matching_predicates = [pred for pred in self.current_predicates if (token in pred)]

        return {'matches' : matching_predicates,
                'cursor_end' : cursor_pos,
                'cursor_start' : cursor_pos - len(token),
                'metadata' : {},
                'status' : 'ok'}


    def do_inspect(self, code, cursor_pos, detail_level=0, omit_sections=()):
        """
        Inspection is supported for the predicates from module jupyter.
        By overriding this method, inspection for further predicates can be implemented.
        """
        token, data = self.get_token_and_jupyter_predicate_inspection_data(code, cursor_pos)

        if data == {}:
            found = False
        else:
            found = True

        return {'status': 'ok', 'data': data, 'metadata': {}, 'found': found}


    def get_token_and_jupyter_predicate_inspection_data(self, code, cursor_pos):
        token = self.get_current_token(self.inspection_splitter, code, cursor_pos)

        if not token:
            # There is no token which can be inspected
            data = {}
        else:
            # Find all matching predicate inspection data
            # If a key of the dictionary contains the current token, the element is assumed to match
            matching_predicate_data = {pred:self.jupyter_predicate_docs[pred] for pred in self.jupyter_predicate_docs if (token in pred)}

            if len(matching_predicate_data) == 0:
                # There is no matching predicate
                data = {}
            else:
                 # Compute plain text and markdown output for the matching predicate data
                jupyter_docs_plain = ''
                jupyter_docs_md = ''
                for pred, data in matching_predicate_data.items():
                    jupyter_docs_plain += data + '\n\n'
                    jupyter_docs_md += '<pre>' + data.replace('\n', '<br>').replace('$', '&#36;') + '<br><br></pre>'

                data = {'text/plain': jupyter_docs_plain, 'text/markdown': jupyter_docs_md}

        return token, data


    def get_current_token(self, splitter, code, cursor_pos):
        if cursor_pos is None:
            cursor_pos = len(code)
        # Get the line where the cursor is and the character offset of the start of the line
        line, offset = line_at_cursor(code, cursor_pos)
        line_cursor = cursor_pos - offset

        # Get the current token in the line
        return splitter.split_line(line, line_cursor)


    ############################################################################
    # Handle server requests and responses
    ############################################################################


    def server_request(self, id, method, params=None, log_response=True):
        """
        Sends a request to the Logtalk server, reads the JSON response, deserializes and returns it.

        If something goes wrong, raises an exception so that an error response is sent to the frontend.

        Raises
        ------
        JSONDecodeError if the response could not be deserialized
        """
        # Create a JSON-RCP Request object (https://www.jsonrpc.org/specification#request_object)
        if params is None:
            request = json.dumps({'jsonrpc':'2.0', 'id':id, 'method':method})
        else:
            request = json.dumps({'jsonrpc':'2.0', 'id':id, 'method':method, 'params':params})
        self.logger.debug('The Request object is: ' + str(request))

        # Send the request to the Logtalk server
        self.logtalk_proc.stdin.write(request)
        self.logtalk_proc.stdin.write('\n')
        self.logtalk_proc.stdin.flush()

        # Read the JSON-RCP Response object (http://www.jsonrpc.org/specification#response_object)
        response_string = self.logtalk_proc.stdout.readline()
        if log_response:
            self.logger.debug('response: ' + response_string)

        try:
            return json.loads(response_string)
        except json.decoder.JSONDecodeError as exception:
            self.logger.debug('The Response object is no valid JSON object: ' + str(response_string))
            self.logger.debug(len(str(response_string)))
            self.logger.debug(ord(str(response_string)))
            raise


    def handle_success_response(self, response_dict):
        """
        Handles a success response by computing output for each term result and sending it to the frontend.

        The dictionary response_dict contains the key 'result'.
        The corresponding value contains the results of the Logtalk terms read from the cell.
        These are given as a dictionary where the keys are integers starting from 1.

        Each of the results is a dictionary with a status member which is either 'halt', 'success', or 'error'.
        For each result, a corresponding response is sent to the client.

        In case of CLPFD a variable might not have been assigned a single value, but a domain instead.
        In that case, the dictionary value of the variable is a dictionary where the value of 'dom' is a string representing the domain.
        If there is a single value, the binding is displayed by the frontend with a '=' (e.g. M = 1).
        Otherwise, it is displayed with 'in' (e.g. X in 1..3).

        Example
        ------
        For the cell code
          "member(M, [1,2,3]), print(M)."
        the result dictionary is:
        {
          "1": {
            "status": "success",
            "type": "query",
            "bindings": {
              "M": "1"
            },
            "output": ""
          }
        }
        The following variable bindings are sent to the frontend:
          M = 1

        CLPFD Example
        ------
        For the cell code
          "X in 1..7, Y #= X+X, Y #\= 3."
        the result dictionary is:
        {
          "1": {
            "status": "success",
            "type": "query",
            "bindings": {
              "X": {
                "dom": "1..7"
              },
              "Y": {
                "dom": "{2}\\/(4..14)"
              }
            },
            "output": ""
          }
        }
        The following variable bindings are sent to the frontend:
          X in 1..7,
          Y in {2}\/(4..14)
        """

        # Read the term results
        result = response_dict["result"]

        if result is None or result == '':
            return {
                'status': 'ok',
                'execution_count': self.kernel.execution_count,
                'payload': [],
                'user_expressions': {},
            }

        is_error = False
        first_error_object = None

        index = 1
        while str(index) in result:
            term_result = result[str(index)]
            status = term_result["status"]

            if status == "halt":
                # The Logtalk server was stopped, so it has to be restarted the next time code is to be executed
                self.kill_logtalk_server()
                self.send_response_display_data(self.backend_data["informational_prefix"] + 'Successfully halted')
            elif status == "error":
                is_error = True
                error_object = self.handle_error_response(term_result)
                if not first_error_object:
                    first_error_object = error_object
            elif status == "success":
                # Handle any additional data and check if the handling was successful
                additional_data_error_keys = self.handle_additional_data(term_result)
                is_error = is_error or additional_data_error_keys
                if is_error and not first_error_object:
                    first_error_object = {
                        'status' : 'error',
                        'ename' : 'error',
                        'evalue' : '',
                        'traceback' : ['The handling of additional data failed for ' + ", ".join(additional_data_error_keys)],
                    }

                # Send the output to the frontend
                output = term_result["output"]
                if output != "":
                    self.send_response_display_data(str(output))

                if term_result["type"] == "query":
                    ansi_escape_codes = ""
                    # Send the variable names and values or a success or failure response to the frontend
                    bindings = term_result["bindings"]
                    if bindings == {}:
                        if additional_data_error_keys:
                            # The handling of the additional data has failed
                            response_text = self.backend_data["failure_response"]
                            ansi_escape_codes = "\x1b[31m" # red
                        else:
                            response_text = self.backend_data["success_response"]
                    else:
                        # Read the variable values
                        variable_values = []
                        for var, val in bindings.items():
                            if isinstance(val, dict):
                                if 'dom' in val:
                                    # CLPFD: the variable has not been assigned a single value, but a domain
                                    variable_values.append(str(var) + ' in ' + str(val['dom']))
                            else:
                                variable_values.append(str(var) + ' = ' + str(val))
                        response_text = ",\n".join(variable_values)

                    self.send_response_display_data(response_text, "\x1b[1m" + ansi_escape_codes) # bold
            index = index + 1

        # If at least one of the terms caused an error, an error reply is sent to the client (corresponding to the first error which was encountered)
        if is_error:
            if first_error_object:
                return first_error_object
            else:
                return {
                    'status' : 'error',
                    'ename' : 'error',
                    'evalue' : '',
                    'traceback' : [],
                }
        else:
            return {
                'status': 'ok',
                'execution_count': self.kernel.execution_count,
                'payload': [],
                'user_expressions': {},
            }


    def handle_error_response(self, response_dict):
        """
        Handles an error response by sending an error message to the frontend.

        The dictionary response_dict contains the key 'error'.
        The corresponding value is a dictionary containing the error data.
        The member 'data' can contain members 'logtalk_message (e.g. a more specific error message)
        and 'output' (output of the request before the error occurred).

        Example
        ------
        For the cell code
          "print(test), 3 is 1 + x."
        the error dictionary is:
        {
            "code": -4712,
            "message": "Exception",
            "data": {
                "logtalk_message": "! Type error in argument 2 of (is)/2\n! expected evaluable, but found x/0\n! goal:  3 is 1+x",
                "output": "test"
            }
        }
        """

        error = response_dict["error"]
        error_code = error['code']

        if error['data']:
            self.handle_additional_data(error['data'])

            # Send the output to the client
            if 'output' in error['data']:
                output = error['data']['output']
                self.send_response_display_data(output)
            else:
                output = ''

        if error_code == -4711:
            ename = 'failure'
            if error['data']['logtalk_message'] != '':
                output += '\n' + error['data']['logtalk_message']
                response_text = error['data']['logtalk_message']
            else:
                output += '\n' + self.backend_data["failure_response"]
                response_text = self.backend_data["failure_response"]
        elif error_code == -4712:
            # Exception: "logtalk_message" contains the error message
            ename = 'exception'
            output += '\n' + error['data']['logtalk_message']
            response_text = error['data']['logtalk_message'] + '\n'
        elif error_code == -4715:
            # Unhandled exception: the server needs to be restarted
            ename = 'unhandled exception'
            output += '\n' + error['data']['logtalk_message']
            self.kill_logtalk_server()
            response_text = error['data']['logtalk_message'] + '\n' + self.backend_data["error_prefix"] + 'The Logtalk server needs to be restarted'
        else:
            ename = 'error'
            output += '\n' + error['message']
            response_text = self.backend_data["error_prefix"] + str(error['message']) + '\n'

        self.send_response_display_data(response_text, self.error_ansi_escape_codes)

        return {
           'status' : 'error',
           'ename' : ename,
           'evalue' : '',
           'traceback' : [output], # Needed for nbgrader validation
        }


    def send_response_display_data(self, text, ansi_escape_codes=""):
        """Sends a response to the frontend containing plain text."""

        display_data = {
            'data': {
                'text/plain': ansi_escape_codes + text
            },
            'metadata': {}}
        self.kernel.send_response(self.kernel.iopub_socket, 'display_data', display_data)


    ############################################################################
    # Handling of additional data
    ############################################################################


    def handle_additional_data(self, dict):
        """
        Handles additional data which may be present in the dict.
        Any of the data processing methods should return True if something goes wrong during the handling.
        Returns a list containing the dictionary keys for which the handling did not succeed.
        """
        failure_keys = []

        if 'predicate_atoms' in dict:
            if self.handle_completion_data_update(dict['predicate_atoms']):
                failure_keys.append(['predicate_atoms'])
        if 'print_sld_tree' in dict:
            if self.handle_print_graph(dict['print_sld_tree']):
                failure_keys.append(['print_sld_tree'])
        if 'print_table' in dict:
            if self.handle_print_table(dict['print_table']):
                failure_keys.append(['print_table'])
        if 'print_transition_graph' in dict:
            if self.handle_print_graph(dict['print_transition_graph']):
                failure_keys.append(['print_transition_graph'])
        if 'set_prolog_backend' in dict:
            if self.handle_set_prolog_backend(dict['set_prolog_backend']):
                failure_keys.append(['set_prolog_backend'])

        return failure_keys


    def handle_completion_data_update(self, predicate_atoms):
        """The user requested to update the predicate data used for code completion."""
        self.current_predicates = predicate_atoms


    def handle_print_graph(self, graph_file_content):
        """
        The string graph_file_content corresponds to the content of a file defining a graph.
        It is used to render an svg file with dot, of which the content is then read in and sent to the frontend so that the graph is displayed.

        Example
        ------
        graph_file_content can look like the following
        graph {
            "1" [label="user:pred"]
            "2" [label="user:sub_goal_1"]
            "3" [label="user:sub_goal_1_1"]
            "4" [label="user:sub_goal_1_2"]
            "5" [label="user:sub_goal_2"]
            "6" [label="user:sub_goal_2_1"]
            "7" [label="user:sub_goal_2_2"]
            "1" -- "2"
            "2" -- "3"
            "2" -- "4"
            "1" -- "5"
            "5" -- "6"
            "5" -- "7"
        }
        """

        # Write the content to a file
        f = open("graph.gv", "w")
        f.write(graph_file_content)
        f.close()

        # Render a svg file
        render(engine='dot', format='svg', filepath='graph.gv', outfile='graph.svg').replace('\\', '/')

        # Read the svg file content
        svg_file = open("graph.svg", "r")
        svg_content = svg_file.read()
        svg_file.close()

        # Remove the created files
        remove("graph.gv")
        remove("graph.svg")

        # Send the data to the client
        display_data = {
            'data': {
                'text/plain': graph_file_content,
                'image/svg+xml': svg_content
            },
            'metadata': {}}
        self.kernel.send_response(self.kernel.iopub_socket, 'display_data', display_data)


    def handle_print_table(self, print_table_dict):
        """
        The dictionary print_table_dict contains the members 'ValuesLists' and 'VariableNames'.
        ValuesLists is a list of lists where each of them is used to compute one line of the table.
        VariableNames is a list of strings from which the header of the table is created.

        Example
        ------
        For the cell code
          "print_table((member(Member, [10,20,30]), Square is Member*Member))."
        the variables dictionary is:
          {'ValuesLists': [['10', '100'], ['20', '400'], ['30', '900']], 'VariableNames': ['Member', 'Square']}
        The markdown text sent to the frontend is:
          Member | Square |
          :- | :- |
          10 | 100 |
          20 | 400 |
          30 | 900 |
        """
        values_lists = print_table_dict["ValuesLists"]
        variable_names = print_table_dict["VariableNames"]

        table_header_markdown_string = ""
        table_markdown_string = ""

        # Create a header line with the variable names
        for variable_name in variable_names:
            table_header_markdown_string = table_header_markdown_string + str(variable_name) + " | "
            table_markdown_string = table_markdown_string + ":- | "

        table_markdown_string = table_header_markdown_string  + "\n" + table_markdown_string

        # For each values list, add a markdown table line
        for values_list in values_lists:
            line_markdown_string = ""
            for value in values_list:
                line_markdown_string = line_markdown_string + str(value) + " | "

            table_markdown_string = table_markdown_string + "\n" + line_markdown_string

        display_data = {'data': {'text/plain': table_markdown_string, 'text/markdown': table_markdown_string.replace('$', '\$').replace('~', '\~')}, 'metadata': {}}
        self.kernel.send_response(self.kernel.iopub_socket, 'display_data', display_data)


    def handle_set_prolog_backend(self, prolog_backend):
        """The user requested to change the active Prolog backend, which needs to be handled by the kernel."""
        return self.kernel.change_prolog_backend(prolog_backend)



#############################################################################
#
#  Copyright (c) 2022-2023 Paulo Moura  
#  Copyright (c) 2022 Anne Brecklinghaus, Michael Leuschel, dgelessus
#  SPDX-License-Identifier: MIT
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################


"""
A Logtalk Jupyter kernel communicating with a Logtalk server with JSON-RPC 2.0 messages.
The communication is based on 'jsonrpc_client.py' from SICStus Prolog 4.5.1.

Several Prolog backends are supported. By default, the SWI-Prolog backend is used.
By defining a 'logtalk_kernel_config.py' file, the Prolog backend to be used can be defined.
In addition to providing an backend (the name of the used Logtalk integration script),
further implementation specific data (a dictionary 'backend_data' with the backend as key) can be defined.
This includes the command line arguments with which the Logtalk server can be started.

Additionally, there is the Logtalk predicate 'jupyter::set_prolog_backend(+Backend)' with which the implementation can be changed
(the argument is the name of the used Logtalk integration script).
In order for this to work, the configured 'backend_data' dictionary needs to contain data for more than one Prolog backend.

An example of a configuration file with an explanation of the options and their default values commented out can be found in the current directory.
When defined, this file needs to be present in one of the Jupyter config paths (can be retrieved with 'jupyter --paths') or the current working directory.

The actual kernel code is not implemented by this kernel class itself.
Instead, there is the file 'logtalk_kernel_base_implementation.py' which defines the class 'LogtalkKernelBaseImplementation'.
When the kernel is started, a (sub)object of this class is created.
It handles the starting of and communication with the Logtalk server.
For all requests (execution, shutdown, completion, inspection) the kernel receives, a 'LogtalkKernelBaseImplementation' method is called.
By creating a subclass of this and defining the path to it as 'kernel_backend_path', the actual implementation code can be replaced.

If no such path is defined, the path itself or the defined class is invalid, a default implementation is used instead.
In case of SWI- and SICStus Prolog, the files 'swi_kernel_implementation.py' and 'sicstus_kernel_implementation.py' are used, which can be found in the current directory.
Otherwise, the base implementation from the file 'logtalk_kernel_base_implementation.py' is loaded.

The Logtalk Jupyter kernel is implemented in a way that basically all functionality except the loading of the configuration can easily be overriden.
This is especially useful for extending the kernel for further Prolog backends.
"""


import importlib.util
import logging
import os
import platform
import sys

from inspect import getmembers, isclass
from ipykernel.kernelbase import Kernel
from jupyter_core.paths import jupyter_config_path
from traitlets import Bool, Dict, Unicode
from traitlets.config.loader import ConfigFileNotFound, PyFileConfigLoader

#import logtalk_kernel.swi_kernel_implementation
#import logtalk_kernel.sicstus_kernel_implementation

from logtalk_kernel.logtalk_kernel_base_implementation import LogtalkKernelBaseImplementation


class LogtalkKernel(Kernel):
    kernel_name = 'logtalk_kernel'
    implementation = kernel_name
    implementation_version = '1.0'
    language_info = {
        'name': 'Logtalk',
        'file_extension': '.lgt',
        'mimetype': 'text/x-logtalk',
        'codemirror_mode': 'logtalk',
    }
    banner = kernel_name


    # Define default configuration options

    # If set to True, the logging level is set to DEBUG by the kernel so that Python debugging messages are logged.
    jupyter_logging = Bool(False).tag(config=True)

    # If set to True, a log file is created by the Logtalk server.
    server_logging = Bool(False).tag(config=True)

    # The Prolog backend integration script with which the server is started.
    # It is required that the backend_data dictionary contains an item with the script name.
    if platform.system() == 'Windows':
        #backend = Unicode('eclipselgt.ps1').tag(config=True)
        #backend = Unicode('gplgt.ps1').tag(config=True)
        #backend = Unicode('lvmlgt.ps1').tag(config=True)
        #backend = Unicode('sicstuslgt.ps1').tag(config=True)
        backend = Unicode('swilgt.ps1').tag(config=True)
        #backend = Unicode('tplgt.ps1').tag(config=True)
        #backend = Unicode('yaplgt.ps1').tag(config=True)
    elif 'LOGTALKHOME' in os.environ and 'LOGTALKUSER' in os.environ and os.environ['LOGTALKHOME'] == os.environ['LOGTALKUSER']:
        #backend = Unicode('eclipselgt.sh').tag(config=True)
        #backend = Unicode('gplgt.sh').tag(config=True)
        #backend = Unicode('lvmlgt.sh').tag(config=True)
        #backend = Unicode('sicstuslgt.sh').tag(config=True)
        backend = Unicode('swilgt.sh').tag(config=True)
        #backend = Unicode('tplgt.sh').tag(config=True)
        #backend = Unicode('yaplgt.sh').tag(config=True)
    else:
        #backend = Unicode('eclipselgt').tag(config=True)
        #backend = Unicode('gplgt').tag(config=True)
        #backend = Unicode('lvmlgt').tag(config=True)
        #backend = Unicode('sicstuslgt').tag(config=True)
        backend = Unicode('swilgt').tag(config=True)
        #backend = Unicode('tplgt').tag(config=True)
        #backend = Unicode('yaplgt').tag(config=True)

    # The default program arguments for supported Prolog backends
    default_program_arguments = {
        "eclipselgt": ["eclipselgt",
                "-P",
                "-e", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt.",
                "--", "--quiet"],
        "eclipselgt.sh": ["eclipselgt.sh",
                "-P",
                "-e", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt.",
                "--", "--quiet"],
        "eclipselgt.ps1": ["eclipselgt.ps1",
                "-P",
                "-e", "\"set_logtalk_flag`(report`,off`)`,logtalk_load`(`'logtalk_server/loader.lgt`'`)`,'::'`(jupyter_server`,start`);halt.\"",
                "--%", "--quiet"],
        "gplgt": ["gplgt",
                "--quiet",
                "--entry-goal", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "gplgt.sh": ["gplgt.sh",
                "--quiet",
                "--entry-goal", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "gplgt.ps1": ["gplgt.ps1",
                "--quiet",
                "--entry-goal", "\"set_logtalk_flag`(report`,off`)`,logtalk_load`(`'logtalk_server/loader.lgt`'`)`,'::'`(jupyter_server`,start`);halt\""],
        "lvmlgt": ["lvmlgt",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt."],
        "lvmlgt.sh": ["lvmlgt.sh",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt."],
        "lvmlgt.ps1": ["lvmlgt.ps1",
                "-q",
                "-g", "\"set_logtalk_flag`(report`,off`)`,logtalk_load`(`'logtalk_server/loader.lgt`'`)`,'::'`(jupyter_server`,start`);halt.\""],
        "sicstuslgt": ["sicstuslgt",
                "--noinfo",
                "--goal", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt.",
                "--nologo"],
        "sicstuslgt.sh": ["sicstuslgt.sh",
                "--noinfo",
                "--goal", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt.",
                "--nologo"],
        "sicstuslgt.ps1": ["sicstuslgt.ps1",
                "--noinfo",
                "--goal", "\"set_logtalk_flag`(report`,off`)`,logtalk_load`(`'logtalk_server/loader.lgt`'`)`,`'::`'`(jupyter_server`,start`);halt.\"",
                "--nologo"],
        "swilgt": ["swilgt",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "swilgt.sh": ["swilgt.sh",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "swilgt.ps1": ["swilgt.ps1",
                "-q",
                "-g", "\"set_logtalk_flag`(report`,off`)`,logtalk_load`(`'logtalk_server/loader.lgt`'`)`,'::'`(jupyter_server`,start`);halt\""],
        "tplgt": ["tplgt",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "tplgt.sh": ["tplgt.sh",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "tplgt.ps1": ["tplgt.ps1",
                "-q",
                "-g", "\"set_logtalk_flag`(report`,off`)`,logtalk_load`(`'logtalk_server/loader.lgt`'`)`,'::'`(jupyter_server`,start`);halt\""],
        "yaplgt": ["yaplgt",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "yaplgt.sh": ["yaplgt.sh",
                "-q",
                "-g", "set_logtalk_flag(report,off),logtalk_load('logtalk_server/loader.lgt'),'::'(jupyter_server,start);halt"],
        "yaplgt.ps1": ["yaplgt.ps1",
                "-q",
                "-g", "\"set_logtalk_flag`(report`,off`)`,logtalk_load`(`'logtalk_server/loader.lgt`'`)`,'::'`(jupyter_server`,start`);halt\""],
    }

    # The implementation specific data which is needed to run the Logtalk server for code execution.
    # This is required to be a dictionary containing at least an entry for the configured backend.
    # Each entry needs to define values for
    # - "failure_response": The output which is displayed if a query fails
    # - "success_response": The output which is displayed if a query succeeds without any variable bindings
    # - "error_prefix": The prefix output for error messages
    # - "informational_prefix": The prefix output for informational messages
    # - "program_arguments": The command line arguments (a list of strings) with which the Logtalk server can be started
    # Additionally, a "kernel_backend_path" can be provided, which needs to be an absolute path to a Python file.
    # The corresponding module is required to define a subclass of LogtalkKernelBaseImplementation named LogtalkKernelImplementation.
    # This can be used to override some of the kernel's basic behavior.
    backend_data = Dict({
        "eclipselgt": {
            "failure_response": "No",
            "success_response": "Yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "eclipselgt.sh": {
            "failure_response": "No",
            "success_response": "Yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "eclipselgt.ps1": {
            "failure_response": "No",
            "success_response": "Yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "gplgt": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "gplgt.sh": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "gplgt.ps1": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "lvmlgt": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "lvmlgt.sh": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "lvmlgt.ps1": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "sicstuslgt": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "sicstuslgt.sh": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "sicstuslgt.ps1": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "swilgt": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "swilgt.sh": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "swilgt.ps1": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "tplgt": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "tplgt.sh": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "tplgt.ps1": {
            "failure_response": "false",
            "success_response": "true",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "yaplgt": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "yaplgt.sh": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        },
        "yaplgt.ps1": {
            "failure_response": "no",
            "success_response": "yes",
            "error_prefix": "!     ",
            "informational_prefix": "% ",
            "program_arguments": "default"
        }
    }).tag(config=True)

    # The keys which are required for each entry in the backend_data dict.
    required_backend_data_keys = [
        "failure_response",
        "success_response",
        "error_prefix",
        "informational_prefix",
        "program_arguments"
    ]

    logger = None

    # A dictionary with implementation ids as keys and the corresponding LogtalkKernelBaseImplementation as value.
    # When a Prolog backend is started, it is added to the dictionary.
    # On kernel shutdown or interruption, all implementations are shutdown/interrupted.
    active_kernel_implementations = {}


    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Set the logging format
        logging.basicConfig(format='[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s')

        self.logger = logging.getLogger()
        # For development, the logging level can be set to level DEBUG, so that all debug messages (including the ones about loading a configuration file) are output
        #self.logger.setLevel(logging.DEBUG)

        # Load the configuration and configured implementation specific data
        self.load_config_file()
        load_exception_message = self.load_backend_data(self.backend)
        if load_exception_message:
            # The configured backend_data is invalid
            raise Exception(load_exception_message)

        # Create an implementation object which starts the Logtalk server
        self.load_kernel_implementation()


    def load_config_file(self):
        """
        Searches the search paths for Jupyter config files and the current working directory for logtalk_kernel_config.py files.
        If such files exist, they are loaded.
        If a config file exists in the current working directory, its configuration overrides values from other config files.

        Based on _load_config_files(cls, basefilename, path, log, raise_config_file_errors) from traitlets.config.application.Application.
        """

        config_file_name = 'logtalk_kernel_config.py'

        # Get the search path for Jupyter config files and add the current working directory
        config_paths = jupyter_config_path()
        config_paths.insert(0, os.getcwd())

        # List all existing Prolog kernel config files
        existing_file_paths = [p for p in config_paths if os.path.exists(os.path.join(p, config_file_name))]
        # existing_file_paths list is in descending priority order, so it needs to be reversed
        existing_file_paths.reverse()

        if not existing_file_paths:
            self.logger.debug("No " + config_file_name + " file found in one of these paths: " + str(config_paths))
            self.logger.debug("Using the default configuration")
            return

        # For all paths, load the config file and update the configuration
        for existing_file_path in existing_file_paths:
            loader = PyFileConfigLoader(config_file_name, path=existing_file_path, log=self.logger)

            config = None
            try:
                config = loader.load_config()
            except ConfigFileNotFound:
                self.logger.error("Could not find the config file " + os.path.join(existing_file_path, config_file_name), exc_info=True)
            except Exception:
                self.logger.error("Exception while loading config file " + os.path.join(existing_file_path, config_file_name), exc_info=True)
            else:
                # Update the configuration
                self.update_config(config)

                # If configured, enable the logging of debug messages
                if self.jupyter_logging:
                    self.logger.setLevel(logging.DEBUG)

                self.logger.debug("Loaded config file: " + loader.full_filename)


    def load_backend_data(self, backend):
        """
        Tries to set the implementation data for the Prolog backend with ID backend.
        If no such data is provided for the given backend ID or the dictionary does not contain all required keys, a corresponding message is returned.
        """

        # Check if there is an item for the backend
        if not backend in self.backend_data:
            return "There is no configured backend_data entry for the Prolog backend '" + backend + "'"

        # Check if all required keys are contained in the dictionary
        missing_keys = []
        for key in self.required_backend_data_keys:
            if not key in self.backend_data[backend]:
                missing_keys.append(key)

        if missing_keys == []:
            # The implementation data is valid
            self.active_backend_data = self.backend_data[backend]
        elif len(missing_keys) == 1:
            return "The configured backend_data dict for the Prolog backend '" + backend + "' needs to contain an entry for '" + missing_keys[0] + "'"
        else:
            return "The configured backend_data dict for the Prolog backend '" + backend + "' needs to contain entries for '" + "', '".join(missing_keys) + "'"


    def load_kernel_implementation(self):
        """
        In order for the kernel to be able to execute code, a (sub)object of 'LogtalkKernelBaseImplementation' is needed.
        If the configured backend_data contains an entry for 'kernel_backend_path', tries to load the corresponding module and create a 'LogtalkKernelImplementation' defined in it.
        This causes the Logtalk server to be started so that code can be executed.

        If no 'kernel_backend_path' is given or it is invalid, a default implementation is used instead.
        For the Prolog backends with ID 'swi' or 'sicstus', there is a module defining the class 'LogtalkKernelImplementation' in the current directory.
        Otherwise, the 'LogtalkKernelBaseImplementation' is used.
        """

        use_default = False

        if 'kernel_backend_path' in self.active_backend_data:
            file_path = self.active_backend_data['kernel_backend_path']

            if not os.path.exists(file_path):
                use_default = True
                self.logger.debug("The configured kernel_backend_path '" + str(file_path) + "' does not exist")
            else:
                self.logger.debug("Loading kernel specific code from '" + str(file_path) + "'")
                # Load the module from the specified file
                (module_name, file_extension)= os.path.splitext(os.path.basename(file_path))
                spec = importlib.util.spec_from_file_location(module_name, file_path)
                kernel_implementation_module = importlib.util.module_from_spec(spec)
                sys.modules[module_name] = kernel_implementation_module
                spec.loader.exec_module(kernel_implementation_module)

                # Try to get the class with name 'LogtalkKernelImplementation' and check if it is valid
                implementation_classes = list(class_pair[1] for class_pair in getmembers(kernel_implementation_module, isclass) if class_pair[0]=='LogtalkKernelImplementation')
                if len(implementation_classes) == 0:
                    use_default = True
                    self.logger.debug("The module at the configured kernel_backend_path needs to define the class 'LogtalkKernelImplementation'")
                else:
                    # Try loading the specific implementation
                    try:
                        self.active_kernel_implementation = kernel_implementation_module.LogtalkKernelImplementation(self)
                        if not isinstance(self.active_kernel_implementation, kernel_implementation_module.LogtalkKernelBaseImplementation):
                            use_default = True
                            self.logger.debug("The class 'LogtalkKernelImplementation' needs to be a subclass of 'LogtalkKernelBaseImplementation'")
                    except Exception:
                        use_default = True
                        self.logger.debug("Exception while creating a 'LogtalkKernelImplementation' object" , exc_info=True)
        else:
            use_default = True
            self.logger.debug('No kernel_backend_path configured')

        if use_default:
            # The configured implementation could not be loaded
            # A default implementation is used instead
#            if self.backend == 'swi':
#                self.logger.debug("Using the default implementation for SWI-Prolog")
#                self.active_kernel_implementation = logtalk_kernel.swi_kernel_implementation.LogtalkKernelImplementation(self)
#            elif self.backend == 'sicstus':
#                self.logger.debug("Using the default implementation for SICStus Prolog")
#                self.active_kernel_implementation = logtalk_kernel.sicstus_kernel_implementation.LogtalkKernelImplementation(self)
#            else:
                self.logger.debug("Using the base implementation")
                self.active_kernel_implementation = LogtalkKernelBaseImplementation(self)

        # Add the Prolog backend specific implementation class to the dictionary of active implementations
        self.active_kernel_implementations[self.backend] = self.active_kernel_implementation


    def change_prolog_backend(self, prolog_backend):
        """
        Change the Prolog backend to the one with ID prolog_backend.
        If there is a running server for that backend, it is activated.
        Otherwise, the backend-specific data is loaded (which starts a new server) and set as the active one.
        Returns True if something goes wrong and the new backend cannot be used.
        """

        self.logger.debug('Change Prolog backend to ' + str(prolog_backend))

        if prolog_backend in self.active_kernel_implementations:
            # There is a running Logtalk server for the provided implementation ID
            # Make it the active one
            self.backend = prolog_backend
            self.active_kernel_implementation = self.active_kernel_implementations[self.backend]
        else:
            # Try to load the implementation specific data
            load_exception_message = self.load_backend_data(prolog_backend)
            if load_exception_message:
                self.logger.debug(load_exception_message)
                # The configured backend_data is invalid
                # Display an error message
                load_exception_message = self.active_backend_data["error_prefix"] + load_exception_message

                display_data = {
                    'data': {
                        'text/plain': load_exception_message,
                        'text/markdown': '<pre style="' + self.active_kernel_implementation.output_text_style + 'color:red">' + load_exception_message + '</pre>'                    },
                    'metadata': {}}
                self.send_response(self.iopub_socket, 'display_data', display_data)
                return True
            else:
                self.backend = prolog_backend
                # Create an implementation object which starts the Logtalk server
                self.load_kernel_implementation()


    def interrupt_all(self):
        # Interrupting the kernel interrupts the running Logtalk processes, so all of them need to be restarted
        for backend, kernel_implementation in self.active_kernel_implementations.items():
            kernel_implementation.kill_logtalk_server()


    ############################################################################
    # Overriden kernel methods
    ############################################################################


    def do_shutdown(self, restart):
        # Shutdown all active Logtalk servers so that no processes are kept running
        for kernel_implementation in self.active_kernel_implementations.values():
            kernel_implementation.do_shutdown(restart)

        return {'status': 'ok', 'restart': restart}


    def do_execute(self, code, silent, store_history=True, user_expressions=None, allow_stdin=False):
        return self.active_kernel_implementation.do_execute(code, silent, store_history, user_expressions, allow_stdin)


    def do_complete(self, code, cursor_pos):
        return self.active_kernel_implementation.do_complete(code, cursor_pos)


    def do_inspect(self, code, cursor_pos, detail_level=0, omit_sections=()):
        return self.active_kernel_implementation.do_inspect(code, cursor_pos, detail_level, omit_sections)



#############################################################################
#
#  Copyright (c) 2022-2023 Paulo Moura  
#  Copyright (c) 2022 Anne Brecklinghaus, Michael Leuschel, dgelessus
#  SPDX-License-Identifier: MIT
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################


"""
Default SICStus Prolog kernel implementation.

Defines the inspection for other predicates than the ones defined in the module jupyter.
"""


import requests

from bs4 import BeautifulSoup, Tag

from logtalk_kernel.logtalk_kernel_base_implementation import LogtalkKernelBaseImplementation

# When overriding this at a different location, add the location of the logtalk_kernel_base_implementation.py file defining the LogtalkKernelBaseImplementation class to the search path for modules
#sys.path.append('/path/to/kernel/logtalk_kernel/logtalk_kernel')
#from logtalk_kernel_base_implementation import LogtalkKernelBaseImplementation


class LogtalkKernelImplementation(LogtalkKernelBaseImplementation):

    def retrieve_predicate_information(self):
        """When inspecting a built-in predicate, the SICStus Prolog version is needed for retrieving the documentation URLs"""
        try:
            version_response_dict = self.server_request(0, 'version', log_response=False)
            self.sicstus_version = version_response_dict["result"]

            self.predicate_doc_links = self.get_predicate_doc_links()
        except Exception as exception:
            self.logger.error(exception, exc_info=True)

        super().retrieve_predicate_information()


    def get_predicate_doc_links(self):
        """
        Retrieves the links to the documentation of all predicates for the current Prolog version from the Predicate Index page (e.g. https://sicstus.sics.se/sicstus/docs/latest/html/sicstus.html/Predicate-Index.html for the latest version).
        The link texts look like the following: 'zip/0 (built-in):', 'assert/[1,2] (built-in, ref page):'

        Returns a dictionary where the keys are the names and arities of the built-in predicates (the part of the link text before the bracket).
        The values are lists containing dictionaries with elements 'link_text' and 'link'.

        Additionally, the dictionary contains elements where the key is the predicate name and arity of a predicate exported by the Prolog module 'jupyter'.
        For such a predicate the corresponding value is the documentation of the predicate as a string.

        Example
        ------
        The dictionary containing the first built-in element only might look like the following:
        {'!/0': [{'link_text': '!/0 (built-in, ref page)', 'link': 'https://sicstus.sics.se/sicstus/docs/4.5.1/html/sicstus.html/mpg_002dref_002dcut.html#index-_0021_002f0-_0028built_002din_002c-ref-page_0029-1'}]}
        """
        try:
            # Get the html content and parse it
            base_url = 'https://sicstus.sics.se/sicstus/docs/' + self.sicstus_version + '/html/sicstus.html/'

            predicate_index_link = base_url + '/Predicate-Index.html'
            response = requests.get(predicate_index_link, timeout=1)
            soup = BeautifulSoup(response.text, 'html.parser')

            predicate_data = {}

            # The table "index-pl" contains all the predicate links
            # Find the hyperlink objects
            tables = soup.find_all('table', class_='index-pl')
            if len(tables) > 0:
                for link in tables[0].find_all('a'):
                    if len(link.contents) > 0:
                        child = link.contents[0]
                        if isinstance(child, Tag):
                            # Remove the part in brackets
                            predicate_string = child.text.split('(')[0].strip()
                            predicate_link_dict = {'link_text': child.text, 'link': base_url + link.get('href')}
                            if predicate_string in predicate_data:
                                link_dicts = predicate_data[predicate_string]
                                link_dicts.append(predicate_link_dict)
                                predicate_data[predicate_string] = link_dicts
                            else:
                                predicate_data[predicate_string] = [predicate_link_dict]

            return predicate_data
        except Exception as exception:
            self.logger.error(exception, exc_info=True)
            return None


    def do_inspect(self, code, cursor_pos, detail_level=0, omit_sections=()):
        """
        For SICStus Prolog, the website https://sicstus.sics.se/sicstus/docs/latest/html/sicstus.html/Predicate-Index.html lists links to the documentation of predicates.
        When inspecting a token, links for matching predicates are shown preceding the docs for predicates from module jupyter.
        """
        # Get the matching predicates from module jupyter
        token, jupyter_data = self.get_token_and_jupyter_predicate_inspection_data(code, cursor_pos)

        if not token:
            # There is no token which can be inspected
            return {'status': 'ok', 'data': {}, 'metadata': {}, 'found': False}

        # If the links could not be retrieved when starting the kernel, try reading them again
        if self.predicate_doc_links is None:
            self.predicate_doc_links = self.get_predicate_doc_links()

        if self.predicate_doc_links is None and jupyter_data == {}:
            # There is no matching predicate
            return {'status': 'ok', 'data': {}, 'metadata': {}, 'found': False}

        # Find the matching predicate links
        # If a key of the dictionary contains the current token, the element is assumed to match
        matching_predicate_data = {pred:self.predicate_doc_links[pred] for pred in self.predicate_doc_links if (token in pred)}

        found = True

        if len(matching_predicate_data) == 0:
            # There is no matching predicate from the Predicate Index website
            if jupyter_data == {}:
                data = {}
                found = False
            else:
                data = jupyter_data
        else:
            # There is a matching predicate from the Predicate Index website
            # Compute the link texts
            jupyter_docs_plain = ''
            jupyter_docs_md = ''

            for pred, data in matching_predicate_data.items():
                if isinstance(data, list):
                    for link_dict in data:
                        jupyter_docs_plain += '\x1b[0m' + link_dict['link_text'] + ':\n\x1b[0;34m' + link_dict['link'] + '\n\n'
                        jupyter_docs_md += '<pre><a href="' + link_dict['link'] + '">' + link_dict['link_text'] + '</a><br><br></pre>'

            if jupyter_data != {}:
                # Append the jupyter docs
                jupyter_docs_plain += '\x1b[0m' + '_'*80 + '\n\n' + jupyter_data['text/plain']
                jupyter_docs_md += '' + '_'*80 + '<br><br>' + jupyter_data['text/markdown']

            data = {'text/plain': jupyter_docs_plain, 'text/markdown': jupyter_docs_md}

        return {'status': 'ok', 'data': data, 'metadata': {}, 'found': found}



#############################################################################
#
#  Copyright (c) 2022-2023 Paulo Moura  
#  Copyright (c) 2022 Anne Brecklinghaus, Michael Leuschel, dgelessus
#  SPDX-License-Identifier: MIT
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################


import argparse
import json
import logging
import os
import shutil
import sys
import tempfile

from jupyter_client.kernelspec import KernelSpecManager

logger = logging.getLogger(__name__)

KERNELSPEC_FILES = [
    'kernel.js',
]

def get_kernelspec_dir_path():
    """
    Get the path of the kernelspec directory where the static files needed for the installation are.
    This currently only includes the kernel.js file,
    because the kernelspec (kernel.json) is generated dynamically.
    """
    dirname = os.path.dirname(__file__)
    kernelspec_dir_path = os.path.join(dirname, 'kernelspec')
    return kernelspec_dir_path


def create_kernelspec(dest_dir):
    with open(os.path.join(dest_dir, 'kernel.json'), 'w', encoding='utf-8') as f:
        kernel_json = {
            'argv': [sys.executable, '-m', 'logtalk_kernel', '-f', '{connection_file}'],
            'display_name': 'Logtalk',
            'language': 'logtalk',
        }
        json.dump(kernel_json, f, ensure_ascii=False, indent=4)

    kernelspec_dir = get_kernelspec_dir_path()
    for file in KERNELSPEC_FILES:
        shutil.copyfile(os.path.join(kernelspec_dir, file), os.path.join(dest_dir, file))

def main(argv=None):
    logging.basicConfig(
         format='%(levelname)s: %(message)s',
         level=logging.INFO,
     )

    ap = argparse.ArgumentParser()
    ap.add_argument(
        '--user',
        action='store_true',
        help="install to the per-user kernel registry instead of sys.prefix (use if you get permission errors during installation)")
    ap.add_argument(
        '--prefix',
        help="install to the given prefix: PREFIX/share/jupyter/kernels/")
    args = ap.parse_args(argv)

    if not args.user and not args.prefix:
        args.prefix = sys.prefix

    with tempfile.TemporaryDirectory() as temp_dir:
        create_kernelspec(temp_dir)
        KernelSpecManager().install_kernel_spec(temp_dir, 'logtalk_kernel', user=args.user, prefix=args.prefix)

if __name__ == '__main__':
    main()



#############################################################################
#
#  Copyright (c) 2022-2023 Paulo Moura  
#  Copyright (c) 2022 Anne Brecklinghaus, Michael Leuschel, dgelessus
#  SPDX-License-Identifier: MIT
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#############################################################################


import platform
import os

c = get_config()

## If set to True, the logging level is set to DEBUG by the kernel so that Python debugging messages are logged.
# Default:
# c.LogtalkKernel.jupyter_logging = False

## If set to True, a log file is created by the Logtalk server
# Default:
# c.LogtalkKernel.server_logging = False

## The Prolog backend integration script with which the server is started.
# Default:
if platform.system() == 'Windows':
    #c.LogtalkKernel.backend = "eclipselgt.ps1"
    #c.LogtalkKernel.backend = "gplgt.ps1"
    #c.LogtalkKernel.backend = "lvmlgt.ps1"
    #c.LogtalkKernel.backend = "sicstuslgt.ps1"
    c.LogtalkKernel.backend = "swilgt.ps1"
    #c.LogtalkKernel.backend = "tplgt.ps1"
    #c.LogtalkKernel.backend = "yaplgt.ps1"
elif 'LOGTALKHOME' in os.environ and 'LOGTALKUSER' in os.environ and os.environ['LOGTALKHOME'] == os.environ['LOGTALKUSER']:
    #c.LogtalkKernel.backend = "eclipselgt.sh"
    #c.LogtalkKernel.backend = "gplgt.sh"
    #c.LogtalkKernel.backend = "lvmlgt.sh"
    #c.LogtalkKernel.backend = "sicstuslgt.sh"
    c.LogtalkKernel.backend = "swilgt.sh"
    #c.LogtalkKernel.backend = "tplgt.sh"
    #c.LogtalkKernel.backend = "yaplgt.sh"
else:
    #c.LogtalkKernel.backend = "eclipselgt"
    #c.LogtalkKernel.backend = "gplgt"
    #c.LogtalkKernel.backend = "lvmlgt"
    #c.LogtalkKernel.backend = "sicstuslgt"
    c.LogtalkKernel.backend = "swilgt"
    #c.LogtalkKernel.backend = "tplgt"
    #c.LogtalkKernel.backend = "yaplgt"

## The implementation specific data which is needed to run the Logtalk server for code execution.
## This is required to be a dictionary containing at least an entry for the configured backend.
## Each entry needs to define values for
## - "failure_response": The output which is displayed if a query fails
## - "success_response": The output which is displayed if a query succeeds without any variable bindings
## - "error_prefix": The prefix output for error messages
## - "informational_prefix": The prefix output for informational messages
## - "program_arguments": The command line arguments (a list of strings) with which the Logtalk server can be started
##                        For all backends, the default Logtalk server can be used by configuring the string "default"
## Additionally, a "kernel_backend_path" can be provided, which needs to be an absolute path to a Python file.
## The corresponding module is required to define a subclass of LogtalkKernelBaseImplementation named LogtalkKernelImplementation.
## This can be used to override some of the kernel's basic behavior.
# Default:
c.LogtalkKernel.backend_data = {
   "eclipselgt": {
       "failure_response": "No",
       "success_response": "Yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "eclipselgt.sh": {
       "failure_response": "No",
       "success_response": "Yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "eclipselgt.ps1": {
       "failure_response": "No",
       "success_response": "Yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "gplgt": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "gplgt.sh": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "gplgt.ps1": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "lvmlgt": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "lvmlgt.sh": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "lvmlgt.ps1": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "sicstuslgt": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "sicstuslgt.sh": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "sicstuslgt.ps1": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "swilgt": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "swilgt.sh": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "swilgt.ps1": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "tplgt": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "tplgt.sh": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "tplgt.ps1": {
       "failure_response": "false",
       "success_response": "true",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "yaplgt": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "yaplgt.sh": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   },
   "yaplgt.ps1": {
       "failure_response": "no",
       "success_response": "yes",
       "error_prefix": "!     ",
       "informational_prefix": "% ",
       "program_arguments": "default"
   }
}
